/*
	GetDups plugin for Avisynth 2.5 - get unique frame duplicates
	Copyright (c) 2004, A.G. Balakhnin aka Fizick
	bag@hotmail.ru, http://bag.hotmail.ru
	From ver. 0.6 added some new modes by S.P.Gavrish (Turyst04) gsp@bank.gov.ua
	From ver. 0.9 added audio info using for alternative detecting (by Turyst04)

	Used for processing of captured 8 mm (Super-8) films (with projector obturator removed)
	Based on Dub plugin by Donald Graft/Klaus Post.

		This program is free software; you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program; if not, write to the Free Software
		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "windows.h"
//++++Audio added (20.06.2005) ----v
#include "crtdbg.h"
#include "audio.h"
//++++Audio added (20.06.2005) ----^
#include "avisynth.h"
#include "info.h"
#include "stdio.h"

#define BLKSIZE 8
#define NORM (235*BLKSIZE*BLKSIZE)
//#define MAX_COPIES 20
#define MYVERSION "0.91"

#define TOP 1
#define BOTTOM 2

class GetDups : public GenericVideoFilter
{
	double threshold;
	int fade;
	bool chroma;
	bool tff;
	int mode;
	int show;
	int log;
	const char *logfname;
	int audio;		//++++Audio added (20.06.2005)
	int audio_thr;	//++++Audio added (20.06.2005)
	int audio_shift;//++++Audio added (20.06.2005)
	bool debug;

	int last_n;
	int SearchSimilar; // changed from bool to int in v.0.8
	int ncur;
	int nsrc;
	bool getdupl;//+++GSP Added


public:
    GetDups(PClip _child,double _threshold,int _fade,bool _chroma,bool _tff,int _mode,int _show,
			int _log,const char *_logfname,int _audio,int _audio_thr,int _audio_shift,bool _debug,IScriptEnvironment* env) :
	    GenericVideoFilter(_child),threshold(_threshold),fade(_fade),chroma(_chroma),tff(_tff),mode(_mode),show(_show),
		log(_log),logfname(_logfname),audio(_audio),audio_thr(_audio_thr),audio_shift(_audio_shift),debug(_debug)
	{
	char buf[80];
		if (!vi.IsYUY2() && !vi.IsYV12())
			env->ThrowError("GetDups: requires YUY2 or YV12 source");
		if (threshold < 0.0 || threshold > 100.0)
			env->ThrowError("GetDups: threshold out of range (0.0-100.0)");
		if (debug)
		{
			char b[80];
			sprintf(b, "GetDups %s by Fizick & Turyst04, Copyright 2004, 2005\n", MYVERSION);
			OutputDebugString(b);
		}
		if (audio){
		    if (! vi.HasAudio())
				env->ThrowError("GetDups: clip has no audio");
			PrepareAudio();
		}
		last_n = -1;
		if(audio)ncur = 0; else ncur = 1;
		nsrc = -1;
		getdupl=true;  //+++GSP Added: not drop prev frame
		metricPrev=100;//+++GSP Added
		hadMin=false;  //+++GSP Added
		SearchSimilar =  TOP | BOTTOM;

		xblocks = (vi.width+BLKSIZE-1) / BLKSIZE;
		yblocks = (vi.height/2+BLKSIZE-1) / BLKSIZE; // added /2 in v 0.2 for per-fields processing

		sumTop = (unsigned int *) malloc(xblocks * yblocks * sizeof(unsigned int));
		sumBottom = (unsigned int *) malloc(xblocks * yblocks * sizeof(unsigned int));
		if (sumTop == NULL || sumBottom == NULL ) env->ThrowError("GetDups: cannot allocate needed memory");
		//child->SetCacheHints(3,0);//CACHE_RANGE,3);//0);
		flog=NULL;
		if(mode==4){mode=0;mod=4;}
		else mod=0;
		if(log){//+++GSP Added: open log file
			flog=fopen(logfname,"w+");
			if(flog!=NULL){sprintf(buf,"diff%\n");
				//fwrite(buf,strlen(buf),1,flog);			
			}
		}//+++GSP end "if(log)"
	}
		~GetDups(){if (sumTop) free(sumTop); if (sumBottom) free(sumBottom); 
					if(flog!=NULL)fclose(flog);//+++GSP Added
		}
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
private:
	void SaveStatProt(double metric, int ncur);													//+++GSP Added
	void SaveStatProt(double metricTop, double metricBottom, int ncur);							//+++GSP Added
	void SaveStatProt(double rawTop, double rawBottom, double smoothTop, double smoothBottom, int ncur);							//+++GSP Added
	int CalcFrameDiff(int ncur,double *metricTop, double *metricBottom, IScriptEnvironment* env);//+++GSP Added
	PVideoFrame cur, prev, dest, src;
  	int xblocks, yblocks;
	unsigned int *sumTop, highest_sumTop;
	unsigned int *sumBottom, highest_sumBottom;
	double metricPrev;
	double diffTop[4],diffBottom[4];
	int mod;
	int ncur0;	 //+++GSP Added
	bool hadMin; //+++GSP Added
	FILE *flog;  //+++GSP Added
	//++++Audio added (20.06.2005) ----v
	void SaveStatProt(int a_max,int a_min,int a_pow,int ncur);
	void PrepareAudio();	//+++GSP Added
    void GetStatAudioFrame(int frame, IScriptEnvironment* env);
	void GetStatAudioFrame16(int *min, int *max, int *pow);
	void GetStatAudioFrame8 (int *min, int *max, int *pow);
    BYTE* audio_buffer;
    int samples_per_frame;
    int log_mono_samples;
	int min;
	int max;
	int pow;
	bool Get_Out;//added (v.0.91)
	bool low_audio;
	bool hi_audio;
	//bool have_good; //deleted (v.0.91)
	//++++Audio added (20.06.2005) ----^
};

//---------------------------------------------------------------------
PVideoFrame __stdcall GetDups::GetFrame(int n, IScriptEnvironment* env)
{
	char buf[80];
	double metricTop = 100, metricBottom = 100;
	double metricMax;//+++GSP Added
	int a_max, a_min, a_pow;

    //WORD *audioframe_buffer; //++++ audio
	
	// Detect random access and local restart.
	dest = env->NewVideoFrame(vi);

	if (n != last_n + 1)//if not next frame (0 frame or bad siquence of input frame)
	{
		if(audio)ncur=n;else ncur = n-1;//set start value of ncur (frame number from input clip)
		SearchSimilar =  TOP | BOTTOM;
		if(mod==4){// use temporal smoothing
			//for(int ln=0;ln<2;ln++)//calculate 2 new frames (4 point) for smoothing
			//CalcFrameDiff(ncur+ln,&diffTop[ln],&diffBottom[ln],env);
			ncur+=2;
			CalcFrameDiff(ncur,&diffBottom[1],&diffTop[1],env);
		}
	}
	last_n = n;

	if(mod==4){// use temporal smoothing
		if(ncur==1)
			CalcFrameDiff(ncur,&diffBottom[1],&diffTop[1],env); //calculate first 2 frames (4 point) for smoothing
	}
	if (mode == 1)//+++GSP Added: not drop prev frame
	{
		if(!getdupl){
			src = child->GetFrame(ncur0, env);// Copy frame saved on prev step
			getdupl=true;
			goto copy;
		}
	}//+++GSP end "if(mode == 1)"
	if (n == 0 && show>1) 
		{ncur = 0;} // start
	else
	{
		//++++Audio added (20.06.2005)  ----v
		if(ncur >= vi.num_frames-2){//GetDups has finished filtering - put last frame to out
			//env->ThrowError("GetDups finishes filtering");
			src = child->GetFrame(nsrc, env);goto copy;}
		//int have_hi_audio =false;//removed from (v.0.91)
		Get_Out=false;//added (v.0.91)
		//++++Audio added (20.06.2005)  ----^
		while ((ncur < vi.num_frames-1) )
		{
			// Restart the duplicate detection with the current frame.
			//   If duplicate of previouse frame is found, it is returned,
			// and we search a next frame different from its previous. 
			// and so on.
			ncur +=1; // next
			//++++Audio added (20.06.2005)  ----v
			if(audio){
				int have_low_audio=false;
				int have_hi_audio =false;//removed to (v.0.91)
				if(ncur==1)GetStatAudioFrame(0, env);
				a_max=max;a_min=min;a_pow=pow;//for show mode
				if((log>0)&&(ncur>0)) SaveStatProt(a_max,a_min,a_pow,ncur);
				GetStatAudioFrame(ncur, env);//return globals: max; min; pow
				if(low_audio) // && !have_good) //deleted (v.0.91)
					 have_low_audio=true;
				else if(audio==2)have_low_audio=false;//added if(audio==2) (v.0.91)
				if(hi_audio)
					 have_hi_audio=true;
				//else have_hi_audio=false; //deleted (v.0.91)
				if(max<abs(min))max=abs(min);
				//have_good=false; //deleted (v.0.91)
				if(max<audio_thr){//low_audio found
					low_audio=true;
				}
				else{low_audio=false;hi_audio=true;}//it's not low_audio and set hi_audio
				//if(audio){
  					if(have_low_audio && have_hi_audio){
						src = child->GetFrame(ncur+audio_shift,env);
						low_audio=false;
						hi_audio=false;
						//have_good=true; //deleted (v.0.91)
						have_hi_audio=false;
						nsrc=ncur;
						Get_Out=true;//added (v.0.91)
						goto copy;   //proces this frame to OutPut
					}
					else{
						if (show > 1 )//show only (not proces)
						{
							src = child->GetFrame(ncur,env);
							nsrc=ncur;
							goto copy;// show mode, exit loop
						}
						continue;
					}
				//}
			}
			//++++Audio added (20.06.2005)  ----^
			if(mod!=4)CalcFrameDiff(ncur,&metricTop,&metricBottom,env);
			else{//mode==4 - use temporal smoothing
				//we allready have 2 from neaded 3 values
				diffTop[0]=diffTop[1];diffBottom[0]=diffBottom[1];//shift it
				CalcFrameDiff(ncur,&diffBottom[1],&diffTop[1],env);//and calculate only 2-d for using during next while-step
				if(!tff){
					metricTop  = (diffBottom[0]/2+diffTop   [1]+diffBottom[1]/2)/2;//convolution temporal smoothing
					metricBottom=(diffTop   [0]/2+diffBottom[0]+diffTop   [1]/2)/2;
				}else{
					metricBottom=(diffBottom[0]/2+diffTop   [1]+diffBottom[1]/2)/2;//convolution temporal smoothing with changed fieldes
					metricTop  = (diffTop   [0]/2+diffBottom[0]+diffTop   [1]/2)/2;
				}
			}
			switch(log)//+++GSP Added: Get metric Statistic and save it in Log
				{
				case 1://save nearest fieldes difference siquence (1 column)
					if(metricTop>metricBottom) SaveStatProt(metricTop,ncur);
					else					SaveStatProt(metricBottom,ncur);	
				break;
				//case 2,3,4: save nearest frame fieldes difference siquence (2 columns, Top with Top, Bottom with Bottom)
				case 2: //if mode==4 - smoothed value`s siquence, else - raw value`s siquence
					SaveStatProt(metricTop,metricBottom,ncur);//(if mode==4 - smoothed value`s siquence)
				break;	
				case 3: //in all modes - raw value`s siquence (befour smoothed)
					if(mod==4)SaveStatProt(diffTop[1],diffBottom[1],ncur);
					else SaveStatProt(metricTop,metricBottom,ncur);
				break;	
				case 4://save raw and smoothed value`s siquence (4 columns, raw Top with Top, Bottom with Bottom, then smoothed)
					if(mod==4)SaveStatProt(diffTop[1],diffBottom[1],metricTop,metricBottom,ncur);
					else SaveStatProt(metricTop,metricBottom,ncur);//for mode!=4 - we havn`t smoothed value`s siquence
			}//+++GSP - end "switch(log)"
			if (show > 1 )//show only (not proces)
			{
				break;// show mode, exit loop
			}
			if(mode > 0){//+++GSP Added: (sinc16fps) Work without threshold (looking for local min after max)
				//double metricMax;
				if(ncur>0){//ignore 1-st frame
					if(metricTop>metricBottom)  metricMax=metricTop;
					else						metricMax=metricBottom;
					if (metricMax == 0){;} // it is drop, do nothing, next
					else{
						if(metricMax>metricPrev){//prev field pair is good
							metricPrev=metricMax;
							if(hadMin){hadMin=false;nsrc=ncur-1;break;}//current metric up after local min => prev pair is good
						}else{hadMin=true;metricPrev=metricMax;}//current metric down => move to min
					}
				}
			}//+++GSP - end "if(mode > 0)"
			else //mode==0
			// Since v. 0.2, we now search not frames, but fields of frames
			if (SearchSimilar && (metricTop != 0 || metricBottom !=0)) // Search similar and not drop frame - rewrited in v 0.5.2
			{
				if (metricTop < metricBottom) // Top is most similar, check it first - added in v.0.4 
				{
					if ( (metricTop <= threshold) && tff==true && (SearchSimilar & TOP)) // top is similar and is first, and prev was different - changed in v.0.5.2
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur-1;// prev frame has good top and middle bottom
						break;
					}
					else if ( (metricTop <= threshold) && tff==false )// top is equal and is second
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur;// cur frame has good top and middle bottom
						break;
					}
					else if ( (metricBottom <= threshold) && tff==true ) // bottom is equal and is second
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur;// cur frame has good bottom and middle top
						break;
					}
					else if ( (metricBottom <= threshold) && tff==false && (SearchSimilar & BOTTOM)) // bottom is similar and is first, and prev was different- changed in v.0.5.2
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur-1;// prev frame has good bottom and middle top
						break;
					}
				}
				else // Bottom is most similar, check it first
				{
					if ( (metricBottom <= threshold) && tff==true ) // bottom is equal and is second
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur;// cur frame has good bottom and middle top
						break;
					}
					else if ( (metricBottom <= threshold) && tff==false && (SearchSimilar & BOTTOM)) // bottom is similar and is first, and prev was different- changed in v.0.5.2
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur-1;// prev frame has good bottom and middle top
						break;
					}
					else if ( (metricTop <= threshold) && tff==true && (SearchSimilar & TOP)) // top is equal and is first, and prev was different
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur-1;// prev frame has good top and middle bottom
						break;
					}
					else if ( (metricTop <= threshold) && tff==false )// top is equal and is second
					{
						SearchSimilar = 0; // similar found, exit loop, now search different
						nsrc = ncur;// cur frame has good top and middle bottom
						break;
					}
				}
				SearchSimilar = (TOP | BOTTOM); // you have secondly found different frame, search any similar field now - added in  v.0.8 (0.5.2)
			}
			else if (metricTop > threshold)				// changed in  v.0.8
				SearchSimilar = (SearchSimilar | TOP);	// new different  TOP found, now search new similar TOP
			else if (metricBottom > threshold)
				SearchSimilar = (SearchSimilar | BOTTOM); // new different  BOTTOM found, now search new similar BOTTOM
			//end "else //mode==0"
		}	// end of while
	} 

		// if no duplicates, return source
	if ( (show>1 && (n ==0) )|| nsrc<0) //ncur >= vi.num_frames
	{
		if (mode > 0)//+++GSP Added: not drop similar field
		{
			if(getdupl)
			{
				ncur0=n;
				src = child->GetFrame(n, env); //force duplicate frame
				if (mode == 1)getdupl=false;
			}
		}//+++GSP end of "if (mode > 0)"
		else src = child->GetFrame(n, env); 
	}
	else{//last duplicate founded
		if (mode > 0)//+++GSP Added: in *.avs befor GetDups must be SeparateFields() 
		{
			if(getdupl)
			{
				if(mode == 2)//drop 1-st field (then use Bob)
				{
					src = child->GetFrame(nsrc, env);//mode==2 => return 2-nd from 2 similar frames
				}else
				{//not drop 1-st field (then use Weave)
					ncur0=nsrc;// save 2-nd from 2 similar frames (current)
					if(((int)((ncur0+0)/2)*2)==(ncur0+0)){//+0->even
						ncur0--;//change fields
						src = child->GetFrame(nsrc, env); //return Top=fouded (1-st from 2 similar frames)
					}
					else src = child->GetFrame(nsrc-1, env); //return Top=Prev (1-st from 2 similar frames)
					getdupl=false;
				} 
			}
		}
		else //+++GSP end of "if (mode > 0)"
			//if(mod==4)src = child->GetFrame(nsrc+0, env); 
			//else	  
			src = child->GetFrame(nsrc  , env);
	}
copy:	//+++GSP Added
	// Copy  frame
	if (vi.IsYUY2())
	{
				env->BitBlt(dest->GetWritePtr(), dest->GetPitch(),
							src->GetReadPtr(), src->GetPitch(),
							src->GetRowSize(), src->GetHeight());
	}
	else
	{
				env->BitBlt(dest->GetWritePtr(PLANAR_Y),dest->GetPitch(PLANAR_Y),
							src->GetReadPtr(PLANAR_Y),	src->GetPitch(PLANAR_Y),
							src->GetRowSize(PLANAR_Y),	src->GetHeight(PLANAR_Y));
				env->BitBlt(dest->GetWritePtr(PLANAR_U),dest->GetPitch(PLANAR_U),
							src->GetReadPtr(PLANAR_U),	src->GetPitch(PLANAR_U),
							src->GetRowSize(PLANAR_U),	src->GetHeight(PLANAR_U));
				env->BitBlt(dest->GetWritePtr(PLANAR_V),dest->GetPitch(PLANAR_V),
							src->GetReadPtr(PLANAR_V),	src->GetPitch(PLANAR_V),
							src->GetRowSize(PLANAR_V),	src->GetHeight(PLANAR_V));
	}

	if (show >0)
	{
		// Generate show data overlay. 
			sprintf(buf, "GetDups v.%s by Fizick & Turyst04", MYVERSION);
			if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 0, buf);
			else  DrawString(dest, 0, 0, buf);

			if (tff) sprintf(buf, "Top Field First");
			else sprintf(buf, "Bottom Field First");
			if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 1, buf);
			else  DrawString(dest, 0, 1, buf);

			sprintf(buf, "Frame %d", n);
			if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 2, buf);
			else  DrawString(dest, 0, 2, buf);
			
			if(audio){
				if(ncur>0){if(a_max<abs(a_min))a_max=abs(a_min);
					sprintf(buf, "Audio max  =  %6d", a_max);
					if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 3, buf);
					else  DrawString(dest, 0, 3, buf);

					sprintf(buf, "Audio pow  = %7d", a_pow);//changed (v.0.91)
					if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 4, buf);
					else  DrawString(dest, 0, 4, buf);

					if(Get_Out)sprintf(buf, "Get to Out = Yes");//added (v.0.91)
					else       sprintf(buf, "Get to Out = No");
					if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 6, buf);
					else  DrawString(dest, 0, 6, buf);
					Get_Out=false;
				}
			}
			else{
				if(metricTop<10.0) sprintf(buf, "Top fields diff    =  %2.2f%%", metricTop);
				else			   sprintf(buf, "Top fields diff    = %2.2f%%", metricTop);	
				if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 3, buf);
				else  DrawString(dest, 0, 3, buf);

				if(metricBottom<10.0) sprintf(buf, "Bottom fields diff =  %2.2f%%", metricBottom);
				else				  sprintf(buf, "Bottom fields diff = %2.2f%%", metricBottom);
				if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 4, buf);
				else  DrawString(dest, 0, 4, buf);
			}
			
		
			if (show == 1) // process and show info
			{
				sprintf(buf, "Source frame = %d", nsrc);
				if (vi.IsYUY2()) DrawStringYUY2(dest, 0, 5, buf);
				else  DrawString(dest, 0, 5, buf);
			}
	}
	if (debug == true)
	{
		// Generate debug data. 
			sprintf(buf, "GetDups: frame %d: fields diff = %2.2f, %2.2f%%", n, metricTop, metricBottom);
			OutputDebugString(buf);
	}

	// Return the appropriate frame. 
	return dest;
}

//---------------------------------------------------------------------
int GetDups::CalcFrameDiff(int ncur, double *metricBottom, double *metricTop, IScriptEnvironment* env){//+++GSP Added
	unsigned char * cur_ptr, *prev_ptr;
	unsigned char * cur_ptrY, *cur_ptrU, *cur_ptrV, *prev_ptrY, *prev_ptrU, *prev_ptrV;
	int cur_pitch, prev_pitch;
	int cur_pitchY, cur_pitchUV, prev_pitchY, prev_pitchUV;

    const unsigned char *srcp;
    const unsigned char *src0p;

    int row_size, height;
	int row_sizeY, row_sizeUV, heightY, heightUV;

	int i, j, x, y;
 
	int fadeTop = 0;
	int fadeBottom = 0;

			cur = child->GetFrame(ncur, env);
			prev = child->GetFrame(ncur-1, env);
			if (vi.IsYUY2())
			{
				cur_ptr = (unsigned char *) cur->GetReadPtr();
				cur_pitch = cur->GetPitch();
				row_size = cur->GetRowSize();
				height = cur->GetHeight();
				prev_ptr = (unsigned char *) prev->GetReadPtr();
				prev_pitch = prev->GetPitch();
			}
			else
			{
				cur_ptrY = (unsigned char *) cur->GetReadPtr(PLANAR_Y);
				cur_ptrU = (unsigned char *) cur->GetReadPtr(PLANAR_U);
				cur_ptrV = (unsigned char *) cur->GetReadPtr(PLANAR_V);
				cur_pitchY = cur->GetPitch(PLANAR_Y);
				row_sizeY = cur->GetRowSize(PLANAR_Y);
				heightY = cur->GetHeight(PLANAR_Y);
				cur_pitchUV = cur->GetPitch(PLANAR_U);
				row_sizeUV = cur->GetRowSize(PLANAR_U);
				heightUV = cur->GetHeight(PLANAR_U);
				prev_ptrY = (unsigned char *) prev->GetReadPtr(PLANAR_Y);
				prev_ptrU = (unsigned char *) prev->GetReadPtr(PLANAR_U);
				prev_ptrV = (unsigned char *) prev->GetReadPtr(PLANAR_V);
				prev_pitchY = prev->GetPitch(PLANAR_Y);
				prev_pitchUV = prev->GetPitch(PLANAR_U);
			}

		//	 Compare frames block lumas 
			// Clear the block sums. 
 			for (i = 0; i < yblocks; i++)
			{
 				for (j = 0; j < xblocks; j++)
				{
					sumTop[i*xblocks+j] = 0; // top field
					sumBottom[i*xblocks+j] = 0; // bottom field
				}
			}

			// Do the comparison. 
			if (vi.IsYUY2())
			{
				src0p = cur_ptr;
				srcp = prev_ptr;
				// total correction
				fadeTop = 0;
				fadeBottom =0;
				for (y = 0; y < height; y+=2)
				{
					for (x = 0; x < row_size; x+=2)
					{
						fadeTop += (int)srcp[x] - (int)src0p[x];// top field
					}
					src0p += cur_pitch;
					srcp += prev_pitch;
					for (x = 0; x < row_size; x+=2)
					{
						fadeBottom += (int)srcp[x] - (int)src0p[x];// bottom field
					}
					src0p += cur_pitch;
					srcp += prev_pitch;
				}

				fadeTop /= (row_size*height/4); // mean luma difference value
				fadeBottom /= (row_size*height/4); // mean luma difference value
				if (abs(fadeTop) > fade) fadeTop = 0; // max correction exceeded
				if (abs(fadeBottom) > fade) fadeBottom = 0;

				src0p = cur_ptr;
				srcp = prev_ptr;

				for (y = 0; y < height; y+=2)
				{
					for (x = 0; x < row_size; x+=2)
					{
						sumTop[(y/(2*BLKSIZE))*xblocks + x/(2*BLKSIZE)] += abs((int)srcp[x] - (int)src0p[x] - fadeTop);// top field, luma correction
						if (chroma == true )
							sumTop[(y/(2*BLKSIZE))*xblocks + (x+1)/(2*BLKSIZE)] += abs((int)srcp[x+1] - (int)src0p[x+1]);
						
					}
					src0p += cur_pitch;
					srcp += prev_pitch;
					for (x = 0; x < row_size; x+=2)
					{
						sumBottom[(y/(2*BLKSIZE))*xblocks + x/(2*BLKSIZE)] += abs((int)srcp[x] - (int)src0p[x] - fadeBottom);// bottom field, luma correction
						if (chroma == true )
							sumBottom[(y/(2*BLKSIZE))*xblocks + (x+1)/(2*BLKSIZE)] += abs((int)srcp[x+1] - (int)src0p[x+1]);
						
					}
					src0p += cur_pitch;
					srcp += prev_pitch;
				}
			}
			else // YV12
			{

				src0p = cur_ptrY;
				srcp = prev_ptrY;
				fadeTop = 0;
				fadeBottom =0;
				{
					for (y = 0; y < heightY; y+=2)
					{
						for (x = 0; x < row_sizeY; x++)
						{
							fadeTop += (int)srcp[x] - (int)src0p[x]; // top
						}
						srcp += prev_pitchY;
						src0p += cur_pitchY;
						for (x = 0; x < row_sizeY; x++)
						{
							fadeBottom += (int)srcp[x] - (int)src0p[x]; // bottom
						}
						srcp += prev_pitchY;
						src0p += cur_pitchY;
					}
				}
				fadeTop /= (row_sizeY*heightY/2); // mean difference value
				fadeBottom /= (row_sizeY*heightY/2); // mean difference value
				if (abs(fadeTop) > fade) fadeTop = 0;
				if (abs(fadeBottom) > fade) fadeBottom = 0;

				src0p = cur_ptrY;
				srcp = prev_ptrY;
				{
					for (y = 0; y < heightY; y+=2)
					{
						for (x = 0; x < row_sizeY; x++)
						{
							sumTop[(y/(2*BLKSIZE))*xblocks + x/BLKSIZE] += abs((int)srcp[x] - (int)src0p[x] - fadeTop); // top
						}
						srcp += prev_pitchY;
						src0p += cur_pitchY;
						for (x = 0; x < row_sizeY; x++)
						{
							sumBottom[(y/(2*BLKSIZE))*xblocks + x/BLKSIZE] += abs((int)srcp[x] - (int)src0p[x] - fadeBottom ); // bottom
						}
						srcp += prev_pitchY;
						src0p += cur_pitchY;
					}
				}
				if (chroma == true) // YV12
				{
					src0p = cur_ptrU;
					srcp = prev_ptrU;
					fadeTop = 0;
					{
						for (y = 0; y < heightUV; y+=1)
						{
							for (x = 0; x < row_sizeUV; x++)
							{
								fadeTop += (int)srcp[x] - (int)src0p[x]; // top
							}
							srcp += prev_pitchUV;
							src0p += cur_pitchUV;
						}
					}
					fadeTop /= (row_sizeUV*heightUV); // mean difference value
					if (abs(fadeTop) > fade) fadeTop = 0;
					fadeBottom = fadeTop; // the same as Top value for YV12

					src0p = cur_ptrU;
					srcp = prev_ptrU;
					{
						for (y = 0; y < heightUV; y+=1)
						{
							for (x = 0; x < row_sizeUV; x++)
							{
								sumTop[(y/BLKSIZE)*xblocks + 2*x/BLKSIZE] += abs((int)srcp[x] - (int)src0p[x] - fadeTop); //top
								sumBottom[(y/BLKSIZE)*xblocks + 2*x/BLKSIZE] += abs((int)srcp[x] - (int)src0p[x] - fadeTop); //thr same value as Top  for YV12
							}
							srcp += prev_pitchUV;
							src0p += cur_pitchUV;
						}
					}

					src0p = cur_ptrV;
					srcp = prev_ptrV;
					fadeTop = 0;
					{
						for (y = 0; y < heightUV; y+=1)
						{
							for (x = 0; x < row_sizeUV; x++)
							{
								fadeTop += (int)srcp[x] - (int)src0p[x]; // top
							}
							srcp += prev_pitchUV;
							src0p += cur_pitchUV;
						}
					}
					fadeTop /= (row_sizeUV*heightUV); // mean difference value
					if (abs(fadeTop) > fade) fadeTop = 0;
					fadeBottom = fadeTop; // the same as Top value for YV12

					src0p = cur_ptrV;
					srcp = prev_ptrV;
					{
						for (y = 0; y < heightUV; y+=1)
						{
							for (x = 0; x < row_sizeUV; x++)
							{
								sumTop[(y/BLKSIZE)*xblocks + 2*x/BLKSIZE] += abs((int)srcp[x] - (int)src0p[x] - fadeTop);
								sumBottom[(y/BLKSIZE)*xblocks + 2*x/BLKSIZE] += abs((int)srcp[x] - (int)src0p[x] - fadeTop);
							}
							srcp += prev_pitchUV;
							src0p += cur_pitchUV;
						}
					}
				}
			}
			// Now find the BLKSIZExBLKSIZE block that has the greatest difference. 
			highest_sumTop = 0;
			highest_sumBottom = 0;
			for (i = 0; i < yblocks; i++)
			{
				for (j = 0; j < xblocks; j++)
				{
					if (sumTop[i * xblocks + j] > highest_sumTop)
					{
						highest_sumTop = sumTop[i * xblocks + j];
					}
					if (sumBottom[i * xblocks + j] > highest_sumBottom)
					{
						highest_sumBottom = sumBottom[i * xblocks + j];
					}
				}
			}
			// Calculate the percentage difference for the block and store the results. 
			
			if((tff)&&(mod==4))
			{
				*metricTop    = (highest_sumTop    * 100.0) / NORM;
				*metricBottom = (highest_sumBottom * 100.0) / NORM;
			}else{
				*metricTop    = (highest_sumBottom * 100.0) / NORM;  //tff & mod==4
				*metricBottom = (highest_sumTop    * 100.0) / NORM;
			}
			return 0;
}

//---------------------------------------------------------------------
void GetDups::SaveStatProt(double rawTop, double rawBottom, double smoothTop, double smoothBottom, int ncur){
  char buf[200];
  	  if(rawTop < 10.0)	sprintf(buf,"%6d\t0%2.2f",ncur,rawTop);
	  else				sprintf(buf, "%6d\t%2.2f",ncur,rawTop);
	  if(rawBottom < 10.0)	sprintf(&buf[strlen(buf)],"\t0%2.2f",rawBottom);
	  else					sprintf(&buf[strlen(buf)], "\t%2.2f",rawBottom);
	  if(smoothTop < 10.0)	sprintf(&buf[strlen(buf)],"\t0%2.2f",smoothTop);
	  else					sprintf(&buf[strlen(buf)], "\t%2.2f",smoothTop);
	  if(smoothBottom < 10.0)sprintf(&buf[strlen(buf)],"%\t0%2.2f\n",smoothBottom);
	  else					 sprintf(&buf[strlen(buf)], "%\t%2.2f\n",smoothBottom);
	if(flog!=NULL)fwrite(buf,strlen(buf),1,flog);
}

//---------------------------------------------------------------------
void GetDups::SaveStatProt(double metricTop, double metricBottom, int ncur){//+++GSP Added
  char buf[200];
//  if(tff){
	  if(metricTop < 10.0)	sprintf(buf,"%6d\t0%2.2f",ncur,metricTop);
	  else					sprintf(buf, "%6d\t%2.2f",ncur,metricTop);
	  if(metricBottom < 10.0)sprintf(&buf[strlen(buf)],"%\t0%2.2f\n",metricBottom);
	  else					 sprintf(&buf[strlen(buf)], "%\t%2.2f\n",metricBottom);
//  }
//  else{
//	  if(metricBottom < 10.0)	sprintf(buf,"%6d\t0%2.2f",ncur,metricBottom);
//	  else						sprintf(buf, "%6d\t%2.2f",ncur,metricBottom);
//	  if(metricTop < 10.0)sprintf(&buf[strlen(buf)],"%\t0%2.2f\n",metricTop);
//	  else				  sprintf(&buf[strlen(buf)], "%\t%2.2f\n",metricTop);
//  }
	if(flog!=NULL)fwrite(buf,strlen(buf),1,flog);
}

//---------------------------------------------------------------------
void GetDups::SaveStatProt(double metric, int ncur){//+++GSP Added
  char buf[80];
	if(metric < 10.0)	sprintf(buf,"%6d\t0%2.2f\n",ncur,metric);
	else				sprintf(buf, "%6d\t%2.2f\n",ncur,metric);
	if(flog!=NULL)fwrite(buf,strlen(buf),1,flog);
}
//---------------------------------------------------------------------
//++++Audio added (20.06.2005)  ----v
void GetDups::SaveStatProt(int a_max,int a_min,int a_pow,int ncur){
  char buf[80];
	sprintf(buf,"%6d\t%6d\t%6d\t%7d\n",ncur,a_max,a_min,a_pow);
	if(flog!=NULL)fwrite(buf,strlen(buf),1,flog);
}
//---------------------------------------------------------------------
/*
 * GetDups::GetStatAudioFrame16
 * 
 * Calc max, min and power of the 16-bit audio data in audio_buffer. 
 * If the source is stereo, the channels are averaged.
 * 
 * Parameters:
 *   min     A pointer to the min calculated value.
 *   max     A pointer to the max calculated value.
 *   pow     A pointer to the pow calculated value.
 */
void GetDups::GetStatAudioFrame16(int *min, int *max, int *pow)
{
    *min = MAX_SHORT;
    *max = MIN_SHORT;
	*pow =  0;
    BYTE *src = audio_buffer;//[x];
    for (int x = 0; x < samples_per_frame; x++)
    {
        int num_samples = 1 << log_mono_samples;
		int sampl=0;
        while (num_samples--)
        {
            sampl += *(short*)src;
			src += 2;
        }
        sampl >>= log_mono_samples;
		if(sampl>*max)*max=sampl;
		if(sampl<*min)*min=sampl;
		*pow+=abs(sampl);
    }
}


/*
 * GetDups::GetStatAudioFrame8
 * 
 * Calc max, min and power of the 8-bit audio data in audio_buffer. 
 * If the source is stereo, the channels are averaged.
 * 
 * Parameters:
 *   min     A pointer to the min calculated value.
 *   max     A pointer to the max calculated value.
 *   pow     A pointer to the pow calculated value.
 */
void GetDups::GetStatAudioFrame8(int *min, int *max, int *pow)
{
    *min = MAX_SHORT;
    *max = MIN_SHORT;
	*pow =  0;
    BYTE *src = audio_buffer;//[x];
    for (int x = 0; x < samples_per_frame; x++)
    {
        int num_samples = 1 << log_mono_samples;
		int sampl=0;
        while (num_samples--)
        {
            sampl += *(BYTE*)src - 128;
            src++;
        }
        sampl >>= log_mono_samples;
		sampl <<= 8; // * 256 - for the same (as 16-bit audio data) level of audio value
		if(sampl>*max)*max=sampl;
		if(sampl<*min)*min=sampl;
		*pow+=abs(sampl);
    }
}


/*
 * GetDups::GetAudioFrame
 * 
 * Get a pointer to the audioframe corresponding to the given video frame.  If
 * the audioframe already exists in the cache, it is returned immediately.
 * Otherwise it is generated from the raw audio data, stored in the cache, and
 * returned.
 *
 * Parameters:
// *   frame      The frame whose audioframe is needed.
 *   env        A pointer to the IScriptEnvironment.
 * 
 * Returns:
 *   A pointer to the audioframe buffer holding the requested audioframe.
 */
void GetDups::GetStatAudioFrame(int frame, IScriptEnvironment* env)
{
        child->GetAudio(audio_buffer, vi.AudioSamplesFromFrames(frame), samples_per_frame, env);
        if (vi.sample_type==SAMPLE_INT16)
            GetStatAudioFrame16(&min, &max, &pow);
        else
            GetStatAudioFrame8(&min, &max, &pow);
}


//---------------------------------------------------------------------
void GetDups::PrepareAudio(){
    int bytes_per_sample = vi.BytesPerAudioSample();
    samples_per_frame = (int)vi.AudioSamplesFromFrames(1);
    audio_buffer = new BYTE[bytes_per_sample * samples_per_frame];
	log_mono_samples=vi.AudioChannels()-1;
    //sample_ranges = new BYTE*[samples_per_frame];
	low_audio=false; hi_audio=false; //have_good=false; //deleted (v.0.91)
	min=0; max=0; pow=0;

}
//++++Audio added (20.06.2005)  ----^
//---------------------------------------------------------------------
AVSValue __cdecl Create_GetDups(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	double  threshold = 7;
	int fade = 10;
	bool chroma = false;
	int mode = 0;
	int show = 0;
	int log  = 0;
	char *logfname = "c:\\gd_stat.log";//[100]
	//++++Audio added (20.06.2005)  ----v
	int audio =0;		//
	int audio_thr=5000;	//=5000 best for my level of the 8mm-grapher sound and camera microphone sensitivity
	int audio_shift=0;	//=0 for my (Turyst04) position of the micro-switch (generates audio clicks)
	//++++Audio added (20.06.2005)  ----^
	bool debug = false;
	bool tff = args[0].AsClip()->GetVideoInfo().IsTFF(); // default as Avisynth assumes

    return new GetDups(args[0].AsClip(),
		args[1].AsFloat(threshold),		// threshold for duplicate declaration
		args[2].AsInt(fade),			// max fade correction in differencing
		args[3].AsBool(chroma),			// use chroma in differencing
		args[4].AsBool(tff),			// top field first
		args[5].AsInt(mode),			// set mode: =0 - standard; =1 or 2 - for input, made from 16,6 fps stabilized film
		args[6].AsInt(show),			// show info or process
		args[7].AsInt(log),				// output log in file
		args[8].AsString(logfname),		// name of the log file
		args[9].AsInt(audio),			// using audio info: =0 - not used (default); =1 - use audio with 1-splashed sound; =2 - use audio with 2-splashed sound (grapher); =3 - (not work in this ver.) use audio and video control
		args[10].AsInt(audio_thr),		// threshold of audio value (default=5000)
		args[11].AsInt(audio_shift),	// shift of the "good frame" (in frames + or -) from default "best" position 0
		args[12].AsBool(debug),			// debug
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("GetDups","c[threshold]f[fade]i[chroma]b[tff]b[mode]i[show]i[log]i[logfname]s[audio]i[audio_thr]i[audio_shift]i[debug]b", Create_GetDups, 0);
    return 0;
}

